#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "core.h"
#include "threads.h"
#include "sock.h"

#include "ocsp_request.h"
#include "support.h"

#define  OCSPD_DEF_MAX_SIZE 	65535
#define  OCSPD_DEF_MAX_READ	1024
#define  MAX_USEC		1000
#define  WAIT_USEC		50

extern OCSPD_CONFIG *ocspd_conf;

OCSP_REQUEST * ocspd_req_get_socket ( int connfd, OCSPD_CONFIG *ocspd_conf) {

	ssize_t fullsize = 0;
	ssize_t newsize  = 0;
	ssize_t maxsize  = 0;
	ssize_t	cont_len = 0;
	ssize_t full_req_size = 0;

	char*   buf      = NULL;

	int	headers  = 0;
	int	sel_ret	 = 0;

	char    *pnt	 = NULL;
	char 	*pnt_end = NULL;
	char    *req_st  = NULL;

	BIO	*mem	 = NULL;
	OCSP_REQUEST * req = NULL;

	struct  timeval   time_out;
	fd_set  readset;
	long ltemp;

	if (!ocspd_conf)
		return NULL;

	maxsize = ocspd_conf->max_req_size < OCSPD_DEF_MAX_READ?
	    OCSPD_DEF_MAX_SIZE: ocspd_conf->max_req_size;
	buf = (char *) malloc(maxsize + OCSPD_DEF_MAX_READ + 1);

	if (!buf) {
		syslog(LOG_ERR,
		    "ERROR: Cannot allocate request buffer (size=%ld)",
		    (long) maxsize + OCSPD_DEF_MAX_READ + 1);
		return NULL;
	}

	if (fcntl(connfd, F_SETFL, O_NONBLOCK) < 0) {
		syslog(LOG_ERR, "[Tread] ERROR, cannot set no non-waiting\n");
		free(buf);
		return NULL;
	}

	FD_ZERO(&readset);
	full_req_size = 0;

	for (;;) {
		FD_SET(connfd, &readset);
		time_out.tv_sec = ocspd_conf->max_timeout_secs;
		time_out.tv_usec = 0;
		sel_ret = select(connfd + 1, &readset, NULL, NULL, &time_out);

		if (sel_ret < 0) {
			if (errno == EINTR) {
				if (ocspd_conf->debug)
					syslog(LOG_ERR,
					    "ERROR::NET::Select Recoverable %s",
					    strerror(errno));
				continue;
			}

			if (ocspd_conf->debug)
				syslog(LOG_ERR, "ERROR::NET::Select %s",
				    strerror(errno));

			free(buf);
			return NULL;
		}

		if (sel_ret == 0) {
			if (ocspd_conf->debug)
				syslog(LOG_ERR,
				    "ERROR::NET::Socket connection timeout");
			free(buf);
			return NULL;
		}

		if (!FD_ISSET(connfd, &readset))
			continue;		/* Should not happen. */

		newsize = recv(connfd, buf + fullsize, OCSPD_DEF_MAX_READ, 0);

		if (newsize == 0)
			break;			/* Connection closed. */

		if (newsize < 0) {
			if (ocspd_conf->debug) {
				syslog(LOG_ERR,
				    "DEBUG::NETWORK::ERROR in RECV::%s [%d]",
				    strerror(errno), errno);

				if (errno == EWOULDBLOCK) {
					syslog(LOG_ERR,
					    "DEBUG::NETWORK::ERROR::"
					    "EWOULDBLOCK");
					continue;
				}
			}

			break;			/* I/O error. */
		}

		if (ocspd_conf->debug)
			syslog(LOG_ERR, "DEBUG::NETWORK::recv received %ld "
			    "(conn = %d)!", (long) newsize, connfd);

		/* Check for headers */

		fullsize += newsize;
		buf[fullsize] = '\0';

		if (!cont_len &&
		    (pnt = strstr_nocase(buf, "Content-Length: "))) {
			ltemp = 0;

			if ((pnt_end = strstr(pnt, "\r\n")))
				sscanf(pnt + 16, "%ld", &ltemp);

			cont_len = ltemp;

			if (ocspd_conf->debug)
				syslog(LOG_ERR, "DEBUG::Got Content Len [%ld]",
					ltemp);
		}

		if (!headers && 
		    ((pnt = strstr(buf, "\r\n\r\n")) ||
		    (pnt = strstr(buf, "\n\n")))) {

			while (*pnt == '\n' || *pnt == '\r')
				*pnt++ = '\0';

			req_st = pnt;
			headers = (int) (req_st - buf);
			full_req_size = headers + cont_len;
		}

		if (fullsize >= maxsize) {
			/* Max Reading size exceeded */
			syslog(LOG_ERR,
			    "ERROR::Max REQUEST size exceeded [%ld]",
			    (long) maxsize);
			free(buf);
			return NULL;
		}

		if (full_req_size > 0 && full_req_size - fullsize < 1)
			break;
	}

	if (!fullsize) {
		free(buf);
		return NULL;			/* Null request. */
	}

	buf[fullsize] = '\0';

	if (!cont_len) {
		if (ocspd_conf->verbose) 
			syslog(LOG_ERR,
			    "ERROR::No Content-Length in REQ Headers");

		if (ocspd_conf->debug)
			fprintf(stderr, "---BEGIN HEADERS---\n%s\n"
			    "---END HEADERS---\n\n", buf);

		free(buf);
		return NULL;
	}

	if (!headers) {
		headers = fullsize;
		req_st = buf + fullsize;
	}

	if (fullsize - headers != cont_len) {
		if (ocspd_conf->verbose)
			syslog(LOG_ERR, "Error::Request not terminated");

		free(buf);
		return NULL;
	}

	if (fullsize >= 5 && strncmp_nocase(buf, "POST ", 5)) {
		if (ocspd_conf->verbose)
			syslog(LOG_ERR,
			    "ERROR::Request::HTTP method is not POST");

		free(buf);
		return NULL;
	}

	if (ocspd_conf->debug)
		fprintf(stderr,
		    "---BEGIN HEADERS---\n%s\n---END HEADERS---\n\n", buf);

        if (!(mem = BIO_new_mem_buf(req_st, cont_len))) {
		BIO *err = NULL;

		if ((err = BIO_new(BIO_s_file())) != NULL) {
			BIO_set_fp(err,stderr,BIO_NOCLOSE);
			ERR_print_errors(err);
			BIO_free(err);
		}

                syslog(LOG_ERR, "ERROR: Internal memory allocation error!");

		if (ocspd_conf->debug)
			fprintf(stderr, "ERROR::req_st=%p [len %ld]\n",
			    req_st, (long) cont_len);
        } else {
		if ((req = d2i_OCSP_REQUEST_bio(mem, NULL)) == NULL) {
			if (ocspd_conf->debug) {
				syslog(LOG_ERR,
				    "ERROR::Parsing in d2i_ function\n");

				fprintf(stderr, "[len %ld] buf=%p -- req_st=%p "
						"(fullsize %ld - "
						"hd_size = %ld - "
						"rq_size = %ld)\n", 
						(long) cont_len, buf, 
						req_st, (long) fullsize,
						(long) (req_st - buf), (long)
						(fullsize - (req_st - buf)));
			}
		}

		BIO_free(mem);
	}

	free(buf);
	return req;
}
